/*
 *
 *  \file       crc32_tablegen.c
 *  \brief      Generate crc32_table_le.h and crc32_table_be.h
 *
 *  Compiling: gcc -std=c99 -o crc32_tablegen crc32_tablegen.c
 *  Add -DWORDS_BIGENDIAN to generate big endian table.
 *  Add -DLZ_HASH_TABLE to generate lzenctab.h (little endian).
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include <stdio.h>
#include "../api/lzma.h"
#include "../../common/tkintegr.h"


static uint32_t crc32_table[8][256];


static void swap_crc32_table(void)
{
	size_t s, b;

	for (s = 0; s < 8; ++s)
		for (b = 0; b < 256; ++b)
			crc32_table[s][b] = bswap32(crc32_table[s][b]);
}


static void init_crc32_table(void)
{
	uint32_t poly32 = UINT32_C(0xEDB88320);
	size_t s, b, i;

	for (s = 0; s < 8; ++s)
	{
		for (b = 0; b < 256; ++b)
		{
			uint32_t r = s == 0 ? b : crc32_table[s - 1][b];

			for (i = 0; i < 8; ++i)
			{
				if (r & 1)
					r = (r >> 1) ^ poly32;
				else
					r >>= 1;
			}

			crc32_table[s][b] = r;
		}
	}
}


#ifdef LZ_HASH_TABLE
static void print_lz_table1(void)
{
	size_t b;

	printf("const uint32_t lzma_lz_hash_table[256] = {");

	for (b = 0; b < 256; ++b)
	{
		if ((b % 4) == 0)
			printf("\n\t");

		printf("0x%08" PRIX32 "UL", crc32_table[0][b]);

		if (b != 255)
			printf(",%s", (b + 1) % 4 == 0 ? "" : " ");
	}

	printf("\n};\n");
}

static void print_lz_table(void)
{
	printf("#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__\n");
	swap_crc32_table();
	print_lz_table1();
	swap_crc32_table();
	printf("#else\n");
	print_lz_table1();
	printf("#endif\n");
}
#else
static void print_crc32_table1(void)
{
	size_t s, b;

	printf("const uint32_t lzma_crc32_table[8][256] = {\n\t{");

	for (s = 0; s < 8; ++s)
	{
		for (b = 0; b < 256; ++b)
		{
			if ((b % 4) == 0)
				printf("\n\t\t");

			printf("0x%08" PRIX32 "UL", crc32_table[s][b]);

			if (b != 255)
				printf(",%s", (b + 1) % 4 == 0 ? "" : " ");
		}

		if (s == 7)
			printf("\n\t}\n};\n");
		else
			printf("\n\t}, {");
	}
}

static void print_crc32_table(void)
{
	printf("#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__\n");
	swap_crc32_table();
	print_crc32_table1();
	swap_crc32_table();
	printf("#else\n");
	print_crc32_table1();
	printf("#endif\n");
}
#endif


int main(void)
{
	init_crc32_table();

	printf("/* This file has been automatically generated by crc32_tablegen.c. */\n\n");
#ifdef LZ_HASH_TABLE
	print_lz_table();
#else
	print_crc32_table();
#endif

	return 0;
}
