/*
 *
 *  \file       crc64.c
 *  \brief      CRC64 calculation
 *
 *  Calculate the CRC64 using the slice-by-four algorithm. This is the same
 *  idea that is used in crc32_fast.c, but for CRC64 we use only four tables
 *  instead of eight to avoid increasing CPU cache usage.
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 *
 */

#include "check.h"
#include "crcmacro.h"

#ifdef __PUREC__
struct lzma_index_hash_s { int dummy; };
struct lzma_index_s { int dummy; };
#endif


#undef S8
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
#	define A1(x) ((uint8_t)(ll_uint64_high(x) >> 24))
#	define S8(x) ll_uint64_lshift(x, 8)
#else
#	define S8(x) ll_uint64_rshift(x, 8)
#	define A1 A
#endif


/* See the comments in crc32_fast.c. They aren't duplicated here. */
LZMA_API(ll_uint64_t) lzma_crc64(const uint8_t *buf, size_t size, ll_uint64_t crc)
{
	crc = ll_uint64_not(crc);

#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	crc = bswap64(crc);
#endif

	if (size > 4) {
		const uint8_t *limit;

		while ((uintptr_t)(buf) & 3) {
#ifdef __PUREC__
			/* Pure-C has a bug accessing the const table */
			ll_uint64_t tmp = lzma_crc64_table[0][*buf++ ^ A1(crc)];
			crc = ll_uint64_xor(tmp, S8(crc));
#else
			crc = ll_uint64_xor(lzma_crc64_table[0][*buf++ ^ A1(crc)], S8(crc));
#endif
			--size;
		}

		limit = buf + (size & ~(size_t)(3));
		size &= (size_t)(3);

		while (buf < limit) {
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
			const uint32_t tmp = ll_uint64_high(crc) ^ *(const uint32_t *)(buf);
#else
			const uint32_t tmp = crc ^ *(const uint32_t *)(buf);
#endif
			buf += 4;

#ifdef __PUREC__
			/* Pure-C has a bug accessing the const table */
			{
			ll_uint64_t tmp1, tmp2, tmp3, tmp4;
			tmp1 = lzma_crc64_table[3][A(tmp)];
			tmp2 = lzma_crc64_table[2][B(tmp)];
			tmp3 = lzma_crc64_table[1][C(tmp)];
			tmp4 = lzma_crc64_table[0][D(tmp)];
			crc = ll_uint64_xor(tmp1,
				  ll_uint64_xor(tmp2,
			      ll_uint64_xor(S32(crc),
			      ll_uint64_xor(tmp3, tmp4))));
			}
#else
			crc = ll_uint64_xor(lzma_crc64_table[3][A(tmp)],
				  ll_uint64_xor(lzma_crc64_table[2][B(tmp)],
			      ll_uint64_xor(S32(crc),
			      ll_uint64_xor(lzma_crc64_table[1][C(tmp)],
				                  lzma_crc64_table[0][D(tmp)]))));
#endif
		}
	}

	while (size-- != 0)
	{
#ifdef __PUREC__
		/* Pure-C has a bug accessing the const table */
		ll_uint64_t tmp = lzma_crc64_table[0][*buf++ ^ A1(crc)];
		crc = ll_uint64_xor(tmp, S8(crc));
#else
		crc = ll_uint64_xor(lzma_crc64_table[0][*buf++ ^ A1(crc)], S8(crc));
#endif
	}

#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
	crc = bswap64(crc);
#endif

	return ll_uint64_not(crc);
}
