/*
 *
 *  \file       crc64_tablegen.c
 *  \brief      Generate crc64_table_le.h and crc64_table_be.h
 *
 *  Compiling: gcc -std=c99 -o crc64_tablegen crc64_tablegen.c
 *  Add -DWORDS_BIGENDIAN to generate big endian table.
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 *
 */

#include <stdio.h>
#include "../api/lzma.h"
#include "../../common/tkintegr.h"


static ll_uint64_t crc64_table[4][256];


static void swap_crc64_table(void)
{
	size_t s, b;

	for (s = 0; s < 4; ++s)
		for (b = 0; b < 256; ++b)
			crc64_table[s][b] = bswap64(crc64_table[s][b]);
}


static void init_crc64_table(void)
{
	ll_uint64_t poly64 = LL_UINT64_C(0xC96C5795UL, 0xD7870F42UL);
	size_t s, b, i;

	for (s = 0; s < 4; ++s)
	{
		for (b = 0; b < 256; ++b)
		{
			ll_uint64_t r = s == 0 ? b : crc64_table[s - 1][b];

			for (i = 0; i < 8; ++i)
			{
				if (r & 1)
					r = (r >> 1) ^ poly64;
				else
					r >>= 1;
			}

			crc64_table[s][b] = r;
		}
	}
}


static void print_crc64_table1(void)
{
	size_t s, b;

	printf("const ll_uint64_t lzma_crc64_table[4][256] = {\n\t{");

	for (s = 0; s < 4; ++s)
	{
		for (b = 0; b < 256; ++b)
		{
			if ((b % 2) == 0)
				printf("\n\t\t");

			printf("LL_UINT64_C(0x%08" PRIX32 "UL, 0x%08" PRIX32 "UL)", (uint32_t)(crc64_table[s][b] >> 32), (uint32_t)crc64_table[s][b]);

			if (b != 255)
				printf(",%s", (b + 1) % 2 == 0 ? "" : " ");
		}

		if (s == 3)
			printf("\n\t}\n};\n");
		else
			printf("\n\t}, {");
	}
}


static void print_crc64_table(void)
{
	printf("#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__\n");
	swap_crc64_table();
	print_crc64_table1();
	swap_crc64_table();
	printf("#else\n");
	print_crc64_table1();
	printf("#endif\n");
}


int main(void)
{
	init_crc64_table();

	printf("/* This file has been automatically generated by crc64_tablegen.c. */\n\n");
	print_crc64_table();
	return 0;
}
