/*
 *
 *  \file       block_header.c
 *  \brief      Utility functions to handle lzma_block
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "common.h"
#include "index.h"

#undef lzma_vli_is_valid
#define lzma_vli_is_valid(vli) (ll_uint64_le(vli, LZMA_VLI_MAX) || ll_uint64_eq(vli, LZMA_VLI_UNKNOWN))

#ifdef __PUREC__
struct lzma_index_hash_s { int dummy; };
struct lzma_index_s { int dummy; };
#endif


extern LZMA_API(lzma_ret)
lzma_block_compressed_size(lzma_block *block, lzma_vli unpadded_size)
{
	uint32_t container_size;
	lzma_vli compressed_size;
	
	/* Validate everything but Uncompressed Size and filters. */
	compressed_size = lzma_block_unpadded_size(block);
	if (ll_uint64_iszero(compressed_size))
		return LZMA_PROG_ERROR;

	container_size = block->header_size + lzma_check_size(block->check);

	/* Validate that Compressed Size will be greater than zero. */
	if (ll_uint64_le(unpadded_size, ll_uint64_make(0, container_size)))
		return LZMA_DATA_ERROR;

	/*
	 * Calculate what Compressed Size is supposed to be.
	 * If Compressed Size was present in Block Header,
	 * compare that the new value matches it.
	 */
	compressed_size = ll_uint64_sub(unpadded_size, ll_uint64_make(0, container_size));
	if (!ll_uint64_eq(block->compressed_size, LZMA_VLI_UNKNOWN) &&
			!ll_uint64_eq(block->compressed_size, compressed_size))
		return LZMA_DATA_ERROR;

	block->compressed_size = compressed_size;

	return LZMA_OK;
}


LZMA_API(lzma_vli) lzma_block_unpadded_size(const lzma_block *block)
{
	lzma_vli unpadded_size;
	
	/*
	 * Validate the values that we are interested in i.e. all but
	 * Uncompressed Size and the filters.
	 *
	 * NOTE: This function is used for validation too, so it is
	 * essential that these checks are always done even if
	 * Compressed Size is unknown.
	 */
	if (block == NULL || block->version > 1
			|| block->header_size < LZMA_BLOCK_HEADER_SIZE_MIN
			|| block->header_size > LZMA_BLOCK_HEADER_SIZE_MAX
			|| (block->header_size & 3)
			|| !lzma_vli_is_valid(block->compressed_size)
			|| ll_uint64_iszero(block->compressed_size)
			|| (unsigned int)(block->check) > LZMA_CHECK_ID_MAX)
		return ll_uint64_make(0, 0);

	/*
	 * If Compressed Size is unknown, return that we cannot know
	 * size of the Block either.
	 */
	if (ll_uint64_eq(block->compressed_size, LZMA_VLI_UNKNOWN))
		return LZMA_VLI_UNKNOWN;

	/* Calculate Unpadded Size and validate it. */
	unpadded_size = ll_uint64_add(block->compressed_size, ll_uint64_make(0, block->header_size + lzma_check_size(block->check)));

	assert(unpadded_size >= UNPADDED_SIZE_MIN);
	if (ll_uint64_gt(unpadded_size, UNPADDED_SIZE_MAX))
		return ll_uint64_make(0, 0);

	return unpadded_size;
}


LZMA_API(lzma_vli) lzma_block_total_size(const lzma_block *block)
{
	lzma_vli unpadded_size = lzma_block_unpadded_size(block);

	if (!ll_uint64_eq(unpadded_size, LZMA_VLI_UNKNOWN))
		unpadded_size = vli_ceil4(unpadded_size);

	return unpadded_size;
}
