/*
 *
 *  \file       filter_decoder.c
 *  \brief      Filter ID mapping to filter-specific functions
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "fdecode.h"
#include "flcommon.h"
#include "l1decode.h"
#include "l2decode.h"
#include "sdecode.h"
#include "dtdecode.h"

#ifdef __PUREC__
struct lzma_index_hash_s { int dummy; };
struct lzma_index_s { int dummy; };
#endif


typedef struct {
	lzma_filter_coder common;

	/*
	 * Decodes Filter Properties.
	 *
	 * \return     - LZMA_OK: Properties decoded successfully.
	 *             - LZMA_OPTIONS_ERROR: Unsupported properties
	 *             - LZMA_MEM_ERROR: Memory allocation failed.
	 */
	lzma_ret (*props_decode)(
			void **options, const lzma_allocator *allocator,
			const uint8_t *props, size_t props_size);

} lzma_filter_decoder;


static const lzma_filter_decoder decoders[] = {
#ifdef HAVE_DECODER_LZMA1
	{ { LZMA_FILTER_LZMA1_C, lzma_lzma_decoder_init, lzma_lzma_decoder_memusage }, lzma_lzma_props_decode },
#endif
#ifdef HAVE_DECODER_LZMA2
	{ { LZMA_FILTER_LZMA2_C, lzma_lzma2_decoder_init, lzma_lzma2_decoder_memusage }, lzma_lzma2_props_decode },
#endif
#ifdef HAVE_DECODER_X86
	{ { LZMA_FILTER_X86_C, lzma_simple_x86_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_POWERPC
	{ { LZMA_FILTER_POWERPC_C, lzma_simple_powerpc_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_IA64
	{ { LZMA_FILTER_IA64_C, lzma_simple_ia64_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_ARM
	{ { LZMA_FILTER_ARM_C, lzma_simple_arm_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_ARMTHUMB
	{ { LZMA_FILTER_ARMTHUMB_C, lzma_simple_armthumb_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_SPARC
	{ { LZMA_FILTER_SPARC_C, lzma_simple_sparc_decoder_init, NULL }, lzma_simple_props_decode },
#endif
#ifdef HAVE_DECODER_DELTA
	{ { LZMA_FILTER_DELTA_C, lzma_delta_decoder_init, lzma_delta_coder_memusage }, lzma_delta_props_decode },
#endif
};


static const lzma_filter_coder *
decoder_find(lzma_vli id)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(decoders); ++i)
		if (ll_uint64_eq(decoders[i].common.id, id))
			return &decoders[i].common;

	return NULL;
}


extern LZMA_API(lzma_bool)
lzma_filter_decoder_is_supported(lzma_vli id)
{
	return decoder_find(id) != NULL;
}


extern lzma_ret
lzma_raw_decoder_init(lzma_next_coder *next, const lzma_allocator *allocator,
		const lzma_filter *options)
{
	return lzma_raw_coder_init(next, allocator, options, decoder_find, false);
}


extern LZMA_API(lzma_ret)
lzma_raw_decoder(lzma_stream *strm, const lzma_filter *options)
{
	lzma_next_strm_init(lzma_raw_decoder_init, strm, (&strm->internal->next, strm->allocator, options));

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	return LZMA_OK;
}


extern LZMA_API(lzma_mem_t) lzma_raw_decoder_memusage(const lzma_filter *filters)
{
	return lzma_raw_coder_memusage(decoder_find, filters);
}


extern LZMA_API(lzma_ret)
lzma_properties_decode(lzma_filter *filter, const lzma_allocator *allocator,
		const uint8_t *props, size_t props_size)
{
	const lzma_filter_decoder *fd;
	
	/* Make it always NULL so that the caller can always safely free() it. */
	filter->options = NULL;

	fd = (const lzma_filter_decoder *)decoder_find(filter->id);
	if (fd == NULL)
		return LZMA_OPTIONS_ERROR;

	if (fd->props_decode == NULL)
		return props_size == 0 ? LZMA_OK : LZMA_OPTIONS_ERROR;

	return fd->props_decode(
			&filter->options, allocator, props, props_size);
}
