/*
 *
 *  \file       filter_decoder.c
 *  \brief      Filter ID mapping to filter-specific functions
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 *
 */

#include "fencode.h"
#include "flcommon.h"
#include "l1encode.h"
#include "l2encode.h"
#include "sencode.h"
#include "dtencode.h"

#ifdef __PUREC__
struct lzma_index_hash_s { int dummy; };
struct lzma_index_s { int dummy; };
struct lzma_lzma1_encoder_s { int dummy; };
#endif


typedef struct {
	lzma_filter_coder common;

	/*
	 * Calculates the recommended Uncompressed Size for .xz Blocks to
	 * which the input data can be split to make multithreaded
	 * encoding possible. If this is NULL, it is assumed that
	 * the encoder is fast enough with single thread.
	 */
	lzma_mem_t (*block_size)(const void *options);

	/* Tells the size of the Filter Properties field. If options are */
	/* invalid, UINT32_MAX is returned. If this is NULL, props_size_fixed */
	/* is used. */
	lzma_ret (*props_size_get)(uint32_t *size, const void *options);
	uint32_t props_size_fixed;

	/*
	 * Encodes Filter Properties.
	 *
	 * \return     - LZMA_OK: Properties encoded successfully.
	 *             - LZMA_OPTIONS_ERROR: Unsupported options
	 *             - LZMA_PROG_ERROR: Invalid options or not enough
	 *               output space
	 */
	lzma_ret (*props_encode)(const void *options, uint8_t *out);

} lzma_filter_encoder;


static const lzma_filter_encoder encoders[] = {
#ifdef HAVE_ENCODER_LZMA1
	{ { LZMA_FILTER_LZMA1_C, lzma_lzma_encoder_init, lzma_lzma_encoder_memusage }, NULL /* FIXME */, NULL, 5, lzma_lzma_props_encode },
#endif
#ifdef HAVE_ENCODER_LZMA2
	{ { LZMA_FILTER_LZMA2_C, lzma_lzma2_encoder_init, lzma_lzma2_encoder_memusage }, lzma_lzma2_block_size /* FIXME */, NULL, 1, lzma_lzma2_props_encode },
#endif
#ifdef HAVE_ENCODER_X86
	{ { LZMA_FILTER_X86_C, lzma_simple_x86_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_POWERPC
	{ { LZMA_FILTER_POWERPC_C, lzma_simple_powerpc_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_IA64
	{ { LZMA_FILTER_IA64_C, lzma_simple_ia64_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_ARM
	{ { LZMA_FILTER_ARM_C, lzma_simple_arm_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_ARMTHUMB
	{ { LZMA_FILTER_ARMTHUMB_C, lzma_simple_armthumb_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_SPARC
	{ { LZMA_FILTER_SPARC_C, lzma_simple_sparc_encoder_init, NULL }, NULL, lzma_simple_props_size, 0, lzma_simple_props_encode },
#endif
#ifdef HAVE_ENCODER_DELTA
	{ { LZMA_FILTER_DELTA_C, lzma_delta_encoder_init, lzma_delta_coder_memusage }, NULL, NULL, 1, lzma_delta_props_encode },
#endif
};


static const lzma_filter_coder *encoder_find(lzma_vli id)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(encoders); ++i)
		if (ll_uint64_eq(encoders[i].common.id, id))
			return &encoders[i].common;

	return NULL;
}


LZMA_API(lzma_bool) lzma_filter_encoder_is_supported(lzma_vli id)
{
	return encoder_find(id) != NULL;
}


LZMA_API(lzma_ret) lzma_filters_update(lzma_stream *strm, const lzma_filter *filters)
{
	size_t count;
	lzma_filter reversed_filters[LZMA_FILTERS_MAX + 1];
	size_t i;

	if (strm->internal->next.update == NULL)
		return LZMA_PROG_ERROR;

	/* Validate the filter chain. */
	if (lzma_raw_encoder_memusage(filters) == LZMA_MEM_MAX)
		return LZMA_OPTIONS_ERROR;

	/* The actual filter chain in the encoder is reversed. Some things */
	/* still want the normal order chain, so we provide both. */
	count = 1;
	while (!ll_uint64_eq(filters[count].id, LZMA_VLI_UNKNOWN))
		++count;

	for (i = 0; i < count; ++i)
		reversed_filters[count - i - 1] = filters[i];

	reversed_filters[count].id = LZMA_VLI_UNKNOWN;

	return strm->internal->next.update(strm->internal->next.coder,
			strm->allocator, filters, reversed_filters);
}


lzma_ret lzma_raw_encoder_init(lzma_next_coder *next, const lzma_allocator *allocator, const lzma_filter *options)
{
	return lzma_raw_coder_init(next, allocator, options, encoder_find, true);
}


LZMA_API(lzma_ret) lzma_raw_encoder(lzma_stream *strm, const lzma_filter *options)
{
	lzma_next_strm_init(lzma_raw_coder_init, strm, (&strm->internal->next, strm->allocator, options, encoder_find, true));

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_SYNC_FLUSH] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	return LZMA_OK;
}


LZMA_API(lzma_mem_t) lzma_raw_encoder_memusage(const lzma_filter *filters)
{
	return lzma_raw_coder_memusage(encoder_find, filters);
}


lzma_mem_t lzma_mt_block_size(const lzma_filter *filters)
{
	lzma_mem_t max = 0;
	size_t i;

	for (i = 0; !ll_uint64_eq(filters[i].id, LZMA_VLI_UNKNOWN); ++i) {
		const lzma_filter_encoder *const fe
				= (const lzma_filter_encoder *)encoder_find(filters[i].id);
		if (fe->block_size != NULL) {
			const lzma_mem_t size = fe->block_size(filters[i].options);
			if (size == 0)
				return 0;

			if (size > max)
				max = size;
		}
	}

	return max;
}


LZMA_API(lzma_ret) lzma_properties_size(uint32_t *size, const lzma_filter *filter)
{
	const lzma_filter_encoder *const fe = (const lzma_filter_encoder *)encoder_find(filter->id);
	if (fe == NULL) {
		/*
		 * Unknown filter - if the Filter ID is a proper VLI,
		 * return LZMA_OPTIONS_ERROR instead of LZMA_PROG_ERROR,
		 * because it's possible that we just don't have support
		 * compiled in for the requested filter.
		 */
		return ll_uint64_le(filter->id, LZMA_VLI_MAX) ? LZMA_OPTIONS_ERROR : LZMA_PROG_ERROR;
	}

	if (fe->props_size_get == NULL) {
		/* No props_size_get() function, use props_size_fixed. */
		*size = fe->props_size_fixed;
		return LZMA_OK;
	}

	return fe->props_size_get(size, filter->options);
}


LZMA_API(lzma_ret) lzma_properties_encode(const lzma_filter *filter, uint8_t *props)
{
	const lzma_filter_encoder *const fe = (const lzma_filter_encoder *)encoder_find(filter->id);
	if (fe == NULL)
		return LZMA_PROG_ERROR;

	if (fe->props_encode == NULL)
		return LZMA_OK;

	return fe->props_encode(filter->options, props);
}
