/*
 *
 *  \file       lzdecode.h
 *  \brief      LZ out window
 *
 *  Authors:    Igor Pavlov
 *              Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#ifndef LZMA_LZ_DECODER_H
#define LZMA_LZ_DECODER_H

#include "common.h"


typedef struct {
	/* Pointer to the dictionary buffer. It can be an allocated buffer */
	/* internal to liblzma, or it can a be a buffer given by the */
	/* application when in single-call mode (not implemented yet). */
	uint8_t *buf;

	/* Write position in dictionary. The next byte will be written to */
	/* buf[pos]. */
	size_t pos;

	/* Indicates how full the dictionary is. This is used by */
	/* dict_is_distance_valid() to detect corrupt files that would */
	/* read beyond the beginning of the dictionary. */
	size_t full;

	/* Write limit */
	size_t limit;

	/* Size of the dictionary */
	size_t size;

	/* True when dictionary should be reset before decoding more data. */
	bool need_reset;

} lzma_dict;


typedef struct {
	size_t dict_size;
	const uint8_t *preset_dict;
	size_t preset_dict_size;
} lzma_lz_options;


typedef struct {
	/* Data specific to the LZ-based decoder */
	void *coder;

	/* Function to decode from in[] to *dict */
	lzma_ret (*code)(void *coder,
			lzma_dict *__restrict dict, const uint8_t *__restrict in,
			size_t *__restrict in_pos, size_t in_size);

	void (*reset)(void *coder, const void *options);

	/* Set the uncompressed size */
	void (*set_uncompressed)(void *coder, lzma_vli uncompressed_size);

	/* Free allocated resources */
	void (*end)(void *coder, const lzma_allocator *allocator);

} lzma_lz_decoder;


void LZMA_LZ_DECODER_INIT(lzma_lz_decoder *c);


lzma_ret lzma_lz_decoder_init(lzma_next_coder *next,
		const lzma_allocator *allocator,
		const lzma_filter_info *filters,
		lzma_ret (*lz_init)(lzma_lz_decoder *lz,
			const lzma_allocator *allocator, const void *options,
			lzma_lz_options *lz_options));

lzma_mem_t lzma_lz_decoder_memusage(size_t dictionary_size);

void lzma_lz_decoder_uncompressed(void *coder, lzma_vli uncompressed_size);


/*
 * Inline functions
 */

/* Get a byte from the history buffer. */
#define dict_get(dict, distance) ((dict)->buf[(dict)->pos - (distance) - 1 + ((distance) < (dict)->pos ? 0 : (dict)->size)])


/* Test if dictionary is empty. */
#define dict_is_empty(dict) ((dict)->full == 0)


/* Validate the match distance */
#define dict_is_distance_valid(dict, distance) ((dict)->full > (distance))


/* Repeat *len bytes at distance. */
bool lzma_lz_dict_repeat(lzma_dict *dict, uint32_t distance, uint32_t *len);


/* Puts one byte into the dictionary. Returns true if the dictionary was */
/* already full and the byte couldn't be added. */
static inline bool
dict_put(lzma_dict *dict, uint8_t byte)
{
	if (unlikely(dict->pos == dict->limit))
		return true;

	dict->buf[dict->pos++] = byte;

	if (dict->pos > dict->full)
		dict->full = dict->pos;

	return false;
}


/* Copies arbitrary amount of data into the dictionary. */
static inline void
dict_write(lzma_dict *__restrict dict, const uint8_t *__restrict in,
		size_t *__restrict in_pos, size_t in_size,
		size_t *__restrict left)
{
	/*
	 * NOTE: If we are being given more data than the size of the
	 * dictionary, it could be possible to optimize the LZ decoder
	 * so that not everything needs to go through the dictionary.
	 * This shouldn't be very common thing in practice though, and
	 * the slowdown of one extra memcpy() isn't bad compared to how
	 * much time it would have taken if the data were compressed.
	 */
	if (in_size - *in_pos > *left)
		in_size = *in_pos + *left;

	*left -= lzma_bufcpy(in, in_pos, in_size,
			dict->buf, &dict->pos, dict->limit);

	if (dict->pos > dict->full)
		dict->full = dict->pos;
}


#define dict_reset(dict) ((dict)->need_reset = true)

#endif
