/*
 *
 *  \file       ia64.c
 *  \brief      Filter for IA64 (Itanium) binaries
 *
 *  Authors:    Igor Pavlov
 *              Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "spriv.h"

#ifdef __PUREC__
struct lzma_index_hash_s { int dummy; };
struct lzma_index_s { int dummy; };
#endif


static size_t
ia64_code(void *simple,
		uint32_t now_pos, bool is_encoder,
		uint8_t *buffer, size_t size)
{
	static const uint32_t BRANCH_TABLE[32] = {
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		4, 4, 6, 6, 0, 0, 7, 7,
		4, 4, 0, 0, 4, 4, 0, 0
	};
	size_t i;
	int j;
	size_t slot;

	(void)simple;
	for (i = 0; i + 16 <= size; i += 16) {
		const uint32_t instr_template = buffer[i] & 0x1F;
		const uint32_t mask = BRANCH_TABLE[instr_template];
		uint32_t bit_pos = 5;

		for (slot = 0; slot < 3; ++slot, bit_pos += 41)
		{
			ll_uint64_t inst_norm;
			size_t byte_pos;
			int bit_res;
			ll_uint64_t instruction;
			
			if (((mask >> slot) & 1) == 0)
				continue;

			byte_pos = (bit_pos >> 3);
			bit_res = (int)bit_pos & 0x7;
			instruction = ll_uint64_zero;

			for (j = 0; j < 6; ++j)
				instruction = ll_uint64_add(instruction, ll_uint64_lshift(ll_uint64_make(0, buffer[i + j + byte_pos]), (8 * j)));

			inst_norm = ll_uint64_rshift(instruction, bit_res);

			if ((ll_uint64_to_uint32(ll_uint64_rshift(inst_norm, 37)) & 0xF) == 0x5 &&
					(ll_uint64_to_uint32(ll_uint64_rshift(inst_norm, 9)) & 0x7) == 0
					/* && (inst_norm & 0x3F) == 0 */
					)
			{
				uint32_t dest;
				uint32_t src = ll_uint64_to_uint32(ll_uint64_rshift(inst_norm, 13)) & UINT32_C(0xFFFFF);
				src |= (ll_uint64_to_uint32(ll_uint64_rshift(inst_norm, 36)) & 1) << 20;

				src <<= 4;

				if (is_encoder)
					dest = now_pos + (uint32_t)(i) + src;
				else
					dest = src - (now_pos + (uint32_t)(i));

				dest >>= 4;

				inst_norm = ll_uint64_and(inst_norm, ll_uint64_not(ll_uint64_lshift(ll_uint64_make(0, UINT32_C(0x8FFFFF)), 13)));
				inst_norm = ll_uint64_or(inst_norm, ll_uint64_lshift(ll_uint64_make(0, dest & UINT32_C(0xFFFFF)), 13));
				inst_norm = ll_uint64_or(inst_norm, ll_uint64_lshift(ll_uint64_make(0, dest & UINT32_C(0x100000)), (36 - 20)));

				instruction = ll_uint64_and(instruction, ll_uint64_make(0, (1 << bit_res) - 1));
				instruction = ll_uint64_or(instruction, ll_uint64_lshift(inst_norm, bit_res));

				for (j = 0; j < 6; j++)
					buffer[i + j + byte_pos] = (uint8_t)ll_uint64_to_uint32(ll_uint64_rshift(instruction, (8 * j)));
			}
		}
	}

	return i;
}


static lzma_ret
ia64_coder_init(lzma_next_coder *next, const lzma_allocator *allocator,
		const lzma_filter_info *filters, bool is_encoder)
{
	return lzma_simple_coder_init(next, allocator, filters,
			ia64_code, 0, 16, 16, is_encoder);
}


extern lzma_ret
lzma_simple_ia64_encoder_init(lzma_next_coder *next,
		const lzma_allocator *allocator,
		const lzma_filter_info *filters)
{
	return ia64_coder_init(next, allocator, filters, true);
}


extern lzma_ret
lzma_simple_ia64_decoder_init(lzma_next_coder *next,
		const lzma_allocator *allocator,
		const lzma_filter_info *filters)
{
	return ia64_coder_init(next, allocator, filters, false);
}
