/*
 *
 *  \file       lzmainfo.c
 *  \brief      lzmainfo tool for compatibility with LZMA Utils
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "sysdefs.h"
#include <errno.h>

#include "lzma.h"
#include "getopt.h"
#include "tkgettxt.h"
#include "tkprognm.h"
#include "tkexit.h"

#ifdef TUKLIB_DOSLIKE
#	include <fcntl.h>
#  if !defined(__atarist__) && !defined(__TOS__)
#	include <io.h>
#  endif
#endif


static const char *uint64_to_str(ll_uint64_t value, int slot)
{
	static char bufs[128][2];
	char reversebuf[128];
	char *end = reversebuf;
	char *p;
	
	if (ll_uint64_iszero(value))
	{
		strcpy(bufs[slot], "0");
	} else
	{
		while (!ll_uint64_iszero(value))
		{
			ll_uint64_t r;
			ll_uint64_t c;
			c = ll_uint64_divmod(value, ll_uint64_make(0, 10), &r);
			*end++ = (int)ll_uint64_to_uint32(r) + '0';
			value = c;
		}
		p = bufs[slot];
		while (end > reversebuf)
			*p++ = *--end;
		*p = '\0';
	}

	return bufs[slot];
}


static void lzma_attribute((__noreturn__))
help(void)
{
	printf(
_("Usage: %s [--help] [--version] [FILE]...\n"
"Show information stored in the .lzma file header"), progname);

	printf(_(
"\nWith no FILE, or when FILE is -, read standard input.\n"));
	printf("\n");

	printf(_("Report bugs to <%s> (in English or Finnish).\n"),
			PACKAGE_BUGREPORT);
	printf(_("%s home page: <%s>\n"), PACKAGE_NAME, PACKAGE_URL);

	tuklib_exit(EXIT_SUCCESS, true);
}


static void lzma_attribute((__noreturn__))
version(void)
{
	puts("lzmainfo (" PACKAGE_NAME ") " LZMA_VERSION_STRING);
	tuklib_exit(EXIT_SUCCESS, true);
}


/* Parse command line options. */
static void
parse_args(int argc, char **argv)
{
	enum {
		OPT_HELP,
		OPT_VERSION
	};

	static const struct option long_opts[] = {
		{ "help",    no_argument, NULL, OPT_HELP },
		{ "version", no_argument, NULL, OPT_VERSION },
		{ NULL,      0,           NULL, 0 }
	};

	int c;
	while ((c = getopt_long(argc, argv, "", long_opts, NULL)) != -1) {
		switch (c) {
		case OPT_HELP:
			help();

		case OPT_VERSION:
			version();

		default:
			exit(EXIT_FAILURE);
		}
	}

	return;
}


/* Primitive base-2 logarithm for integers */
static uint32_t
my_log2(uint32_t n)
{
	uint32_t e;
	for (e = 0; n > 1; ++e, n /= 2) ;
	return e;
}


/* Parse the .lzma header and display information about it. */
static bool
lzmainfo(const char *name, FILE *f)
{
	uint8_t buf[13];
	lzma_options_lzma *opt;
	size_t size;
	lzma_filter filter;
	ll_uint64_t uncompressed_size;
	int i;

	size = fread(buf, 1, sizeof(buf), f);

	if (size != 13) {
		fprintf(stderr, "%s: %s: %s\n", progname, name,
				ferror(f) ? strerror(errno)
				: _("File is too small to be a .lzma file"));
		return true;
	}

	filter.id = LZMA_FILTER_LZMA1;
	filter.options = NULL;

	/* Parse the first five bytes. */
	switch (lzma_properties_decode(&filter, NULL, buf, 5)) {
	case LZMA_OK:
		break;

	case LZMA_OPTIONS_ERROR:
		fprintf(stderr, "%s: %s: %s\n", progname, name,
				_("Not a .lzma file"));
		return true;

	case LZMA_MEM_ERROR:
		fprintf(stderr, "%s: %s\n", progname, strerror(ENOMEM));
		exit(EXIT_FAILURE);

	default:
		fprintf(stderr, "%s: %s\n", progname,
				_("Internal error (bug)"));
		exit(EXIT_FAILURE);
	}

	/* Uncompressed size */
	uncompressed_size = ll_uint64_zero;
	for (i = 0; i < 8; ++i)
		uncompressed_size = ll_uint64_or(uncompressed_size, ll_uint64_lshift(ll_uint64_make(0, buf[5 + i]), (i * 8)));

	/*
	 * Display the results. We don't want to translate these and also
	 * will use MB instead of MiB, because someone could be parsing
	 * this output and we don't want to break that when people move
	 * from LZMA Utils to XZ Utils.
	 */
	if (f != stdin)
		printf("%s\n", name);

	printf("Uncompressed size:             ");
	if (ll_uint64_eq(uncompressed_size, ll_uint64_max))
		printf("Unknown");
	else
		printf("%s MB (%s bytes)",
				uint64_to_str(ll_uint64_rshift(ll_uint64_add(uncompressed_size, ll_uint64_make(0, UINT32_C(0x80000))), 20), 0),
				uint64_to_str(uncompressed_size, 1));

	opt = filter.options;

	printf("\nDictionary size:               "
			"%" PRIu32 " MB (2^%" PRIu32 " bytes)\n"
			"Literal context bits (lc):     %" PRIu32 "\n"
			"Literal pos bits (lp):         %" PRIu32 "\n"
			"Number of pos bits (pb):       %" PRIu32 "\n",
			(opt->dict_size + UINT32_C(0x80000)) / UINT32_C(0x100000),
			my_log2(opt->dict_size), opt->lc, opt->lp, opt->pb);

	free(opt);

	return false;
}


extern int
main(int argc, char **argv)
{
	int ret = EXIT_SUCCESS;

	tuklib_progname_init(argv);
	tuklib_gettext_init(PACKAGE, LOCALEDIR);

#if defined(LZMA_SLB) && (defined(__atarist__) || defined(__TOS__))
	{
		long _ret;
		if ((_ret = slb_lzma_open(NULL)) < 0)
		{
			fprintf(stderr, "cannot load " LZMA_SHAREDLIB_NAME ": %ld\n", _ret);
			return EXIT_FAILURE;
		}
	}
#endif
	parse_args(argc, argv);

#ifdef TUKLIB_DOSLIKE
#  if !defined(__atarist__) && !defined(__TOS__)
	setmode(fileno(stdin), O_BINARY);
#endif
#endif

	/* We print empty lines around the output only when reading from */
	/* files specified on the command line. This is due to how */
	/* LZMA Utils did it. */
	if (optind == argc) {
		if (lzmainfo("(stdin)", stdin))
			ret = EXIT_FAILURE;
	} else {
		printf("\n");

		do {
			if (strcmp(argv[optind], "-") == 0) {
				if (lzmainfo("(stdin)", stdin))
					ret = EXIT_FAILURE;
			} else {
				FILE *f = fopen(argv[optind], "r");
				if (f == NULL) {
					ret = EXIT_FAILURE;
					fprintf(stderr, "%s: %s: %s\n",
							progname,
							argv[optind],
							strerror(errno));
					continue;
				}

				if (lzmainfo(argv[optind], f))
					ret = EXIT_FAILURE;

				printf("\n");
				fclose(f);
			}
		} while (++optind < argc);
	}

	tuklib_exit(ret, true);
	return ret;
}
