/*
 *
 *  \file       mytime.c
 *  \brief      Time handling functions
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "private.h"
#include <time.h>

#if !(defined(HAVE_CLOCK_GETTIME) && HAVE_DECL_CLOCK_MONOTONIC)
#	include <sys/time.h>
#endif

ll_uint64_t opt_flush_timeout = LL_UINT64_C(0, 0);
bool flush_needed;

static ll_uint64_t start_time;
static ll_uint64_t next_flush;


/*
 *  \brief      Get the current time as milliseconds
 *
 * It's relative to some point but not necessarily to the UNIX Epoch.
 */
static ll_uint64_t mytime_now(void)
{
	/* NOTE: HAVE_DECL_CLOCK_MONOTONIC is always defined to 0 or 1. */
#if defined(HAVE_CLOCK_GETTIME) && HAVE_DECL_CLOCK_MONOTONIC
	/*
	 * If CLOCK_MONOTONIC was available at compile time but for some
	 * reason isn't at runtime, fallback to CLOCK_REALTIME which
	 * according to POSIX is mandatory for all implementations.
	 */
	static clockid_t clk_id = CLOCK_MONOTONIC;
	struct timespec tv;
	while (clock_gettime(clk_id, &tv))
		clk_id = CLOCK_REALTIME;

	return ll_uint64_add(ll_uint64_mul(ll_uint64_make(0, tv.tv_sec), ll_uint64_make(0, 1000)), ll_uint64_make(0, tv.tv_nsec / UINT32_C(1000000)));
#elif defined (__PUREC__)
	return ll_uint64_make(0, clock() * 5);
#else
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return ll_uint64_add(ll_uint64_mul(ll_uint64_make(0, tv.tv_sec), ll_uint64_make(0, 1000)), ll_uint64_make(0, tv.tv_usec / 1000));
#endif
}


void mytime_set_start_time(void)
{
	start_time = mytime_now();
	next_flush = ll_uint64_add(start_time, opt_flush_timeout);
	flush_needed = false;
}


ll_uint64_t mytime_get_elapsed(void)
{
	return ll_uint64_sub(mytime_now(), start_time);
}


void mytime_set_flush_time(void)
{
	next_flush = ll_uint64_add(mytime_now(), opt_flush_timeout);
	flush_needed = false;
}


int mytime_get_flush_timeout(void)
{
	ll_uint64_t now, remaining;

	if (ll_uint64_iszero(opt_flush_timeout) || opt_mode != MODE_COMPRESS)
		return -1;

	now = mytime_now();
	if (ll_uint64_ge(now, next_flush))
		return 0;

	remaining = ll_uint64_sub(next_flush, now);
	return ll_uint64_gt(remaining, ll_uint64_make(0, INT_MAX)) ? INT_MAX : (int)ll_uint64_to_uint32(remaining);
}
