/*
 *
 *  \file       util.c
 *  \brief      Miscellaneous utility functions
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "private.h"
#include <stdarg.h>


/* Buffers for uint64_to_str() and uint64_to_nicestr() */
static char bufs[8][128];

/* Thousand separator support in uint64_to_str() and uint64_to_nicestr() */
static enum { UNKNOWN, WORKS, BROKEN } thousand = UNKNOWN;


void *xrealloc(void *ptr, size_t size)
{
	void *p = ptr;

	/*
	 * Save ptr so that we can free it if realloc fails.
	 * The point is that message_fatal ends up calling stdio functions
	 * which in some libc implementations might allocate memory from
	 * the heap. Freeing ptr improves the chances that there's free
	 * memory for stdio functions if they need it.
	 */
	ptr = realloc(ptr, size);

	if (ptr == NULL) {
		const int saved_errno = errno;
		free(p);
		message_fatal("%s", strerror(saved_errno));
	}

	return ptr;
}


char *xstrdup(const char *src)
{
	size_t size;
	char *dest;
	assert(src != NULL);
	size = strlen(src) + 1;
	dest = xmalloc(size);
	return memcpy(dest, src, size);
}


ll_uint64_t str_to_uint64(const char *name, const char *value, ll_uint64_t min, ll_uint64_t max)
{
	ll_uint64_t result = ll_uint64_zero;

	/* Skip blanks. */
	while (*value == ' ' || *value == '\t')
		++value;

	/* Accept special value "max". Supporting "min" doesn't seem useful. */
	if (strcmp(value, "max") == 0)
		return max;

	if (*value < '0' || *value > '9')
		message_fatal(_("%s: Value is not a non-negative "
				"decimal integer"), value);

	do {
		ll_uint64_t add;
		
		/* Don't overflow. */
		if (ll_uint64_gt(result, ll_uint64_make(0x19999999UL, 0x99999998UL))) /* UINT64_MAX / 10 - 9 */
			goto error;

		result = ll_uint64_mul(result, ll_uint64_make(0, 10));

		add = ll_uint64_make(0, *value - '0');

		result = ll_uint64_add(result, add);
		++value;
	} while (*value >= '0' && *value <= '9');

	if (*value != '\0') {
		/*
		 * Look for suffix. Originally this supported both base-2
		 * and base-10, but since there seems to be little need
		 * for base-10 in this program, treat everything as base-2
		 * and also be more relaxed about the case of the first
		 * letter of the suffix.
		 */
		int shift = 0;
		if (*value == 'k' || *value == 'K')
			shift = 10;
		else if (*value == 'm' || *value == 'M')
			shift = 20;
		else if (*value == 'g' || *value == 'G')
			shift = 30;

		++value;

		/* Allow also e.g. Ki, KiB, and KB. */
		if (*value != '\0' && strcmp(value, "i") != 0
				&& strcmp(value, "iB") != 0
				&& strcmp(value, "B") != 0)
			shift = 0;

		if (shift == 0) {
			message(V_ERROR, _("%s: Invalid multiplier suffix"),
					value - 1);
			message_fatal(_("Valid suffixes are `KiB' (2^10), "
					"`MiB' (2^20), and `GiB' (2^30)."));
		}

		/* Don't overflow here either. */
		if (ll_uint64_gt(result, ll_uint64_rshift(ll_uint64_max, shift)))
			goto error;

		result = ll_uint64_lshift(result, shift);
	}

	if (ll_uint64_lt(result, min) || ll_uint64_gt(result, max))
		goto error;

	return result;

error:
	message_fatal(_("Value of the option `%s' must be in the range [%s, %s]"),
				name, uint64_to_str(min, false, 0), uint64_to_str(max, false, 1));
	return result;
}


ll_uint64_t round_up_to_mib(ll_uint64_t n)
{
	return ll_uint64_rshift(ll_uint64_add(n, ll_uint64_make(0, 0xFFFFFUL)), 20);
}


/* Check if thousand separator is supported. Run-time checking is easiest, */
/* because it seems to be sometimes lacking even on POSIXish system. */
static bool check_thousand_sep(uint32_t slot)
{
	if (thousand == UNKNOWN) {
		bufs[slot][0] = '\0';
		sprintf(bufs[slot], "%'u", 1U);
		thousand = bufs[slot][0] == '1' ? WORKS : BROKEN;
	}
	return thousand == WORKS;
}


const char *uint64_to_str(ll_uint64_t value, bool use_thousands, uint32_t slot)
{
	assert(slot < ARRAY_SIZE(bufs));

	if (use_thousands && check_thousand_sep(slot))
	{
		sprintf(bufs[slot], "%'" PRIu64, value);
	} else
	{
		char reversebuf[128];
		char *end = reversebuf;
		char *p;
		
		if (ll_uint64_iszero(value))
		{
			strcpy(bufs[slot], "0");
		} else
		{
			while (!ll_uint64_iszero(value))
			{
				ll_uint64_t r;
				ll_uint64_t c;
				c = ll_uint64_divmod(value, ll_uint64_make(0, 10), &r);
				*end++ = (int)ll_uint64_to_uint32(r) + '0';
				value = c;
			}
			p = bufs[slot];
			while (end > reversebuf)
				*p++ = *--end;
			*p = '\0';
		}
	}

	return bufs[slot];
}


const char *uint64_to_nicestr(ll_uint64_t value, enum nicestr_unit unit_min,
		enum nicestr_unit unit_max, bool always_also_bytes,
		uint32_t slot)
{
	static const char suffix[5][4] = { "B", "KiB", "MiB", "GiB", "TiB" };
	enum nicestr_unit unit;
	char *pos;

	assert(unit_min <= unit_max);
	assert(unit_max <= NICESTR_TIB);
	assert(slot < ARRAY_SIZE(bufs));

	check_thousand_sep(slot);

	unit = NICESTR_B;
	pos = bufs[slot];

	if ((unit_min == NICESTR_B && ll_uint64_lt(value, ll_uint64_make(0, 10000))) || unit_max == NICESTR_B)
	{
		/* The value is shown as bytes. */
		if (thousand == WORKS)
			pos += sprintf(pos, "%'" PRIu32, ll_uint64_to_uint32(value));
		else
			pos += sprintf(pos, "%" PRIu32, ll_uint64_to_uint32(value));
	} else {
		/* Scale the value to a nicer unit. Unless unit_min and */
		/* unit_max limit us, we will show at most five significant */
		/* digits with one decimal place. */
		double d = (double)ll_uint64_to_size_t(value);
		do {
			d /= 1024.0;
			++unit;
		} while (unit < unit_min || (d > 9999.9 && unit < unit_max));

		if (thousand == WORKS)
			pos += sprintf(pos, "%'.1f", d);
		else
			pos += sprintf(pos, "%.1f", d);
	}

	pos += sprintf(pos, " %s", suffix[unit]);

	if (always_also_bytes && ll_uint64_ge(value, ll_uint64_make(0, 10000))) {
		if (thousand == WORKS)
			sprintf(pos, " (%'" PRIu64 " B)", value);
		else
			sprintf(pos, " (%s B)", uint64_to_str(value, false, 0));
	}

	return bufs[slot];
}


bool is_empty_filename(const char *filename)
{
	if (filename[0] == '\0') {
		message_error(_("Empty filename, skipping"));
		return true;
	}

	return false;
}


bool is_tty_stdin(void)
{
	const bool ret = isatty(STDIN_FILENO);

	if (ret)
		message_error(_("Compressed data cannot be read from "
				"a terminal"));

	return ret;
}


bool is_tty_stdout(void)
{
	const bool ret = isatty(STDOUT_FILENO);

	if (ret)
		message_error(_("Compressed data cannot be written to "
				"a terminal"));

	return ret;
}
