/*
 *
 *  \file       test_index.c
 *  \brief      Tests functions handling the lzma_index structure
 *
 *  Author:     Lasse Collin
 *
 *  This file has been put into the public domain.
 *  You can do whatever you want with this file.
 */

#include "tests.h"

#define MEMLIMIT ll_uint64_make(0x00000000, UINT32_C(0x00100000))	/* 1MiB */

#define SMALL_COUNT 3
#define BIG_COUNT 5555


static lzma_index *create_empty(void)
{
	lzma_index *i = lzma_index_init(NULL);

	expect(i != NULL);
	return i;
}


static lzma_index *create_small(void)
{
	lzma_index *i = lzma_index_init(NULL);

	expect(i != NULL);
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 101), ll_uint64_make(0, 555)) == LZMA_OK);
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 602), ll_uint64_make(0, 777)) == LZMA_OK);
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 804), ll_uint64_make(0, 999)) == LZMA_OK);
	return i;
}


static lzma_index *create_big(void)
{
	lzma_vli total_size = ll_uint64_zero;
	lzma_vli uncompressed_size = ll_uint64_zero;
	uint32_t n;
	size_t j;

	lzma_index *i = lzma_index_init(NULL);

	expect(i != NULL);

	/* Add pseudo-random sizes (but always the same size values). */
	n = 11;
	for (j = 0; j < BIG_COUNT; ++j)
	{
		uint32_t t;

		n = 7019 * n + 7607;
		t = n * 3011;
		expect(lzma_index_append(i, NULL, ll_uint64_make(0, t), ll_uint64_make(0, n)) == LZMA_OK);
		total_size = ll_uint64_add(total_size, ll_uint64_make(0, (t + 3) & ~0x3));
		uncompressed_size = ll_uint64_add(uncompressed_size, ll_uint64_make(0, n));
	}

	expect(ll_uint64_eq(lzma_index_block_count(i), ll_uint64_make(0, BIG_COUNT)));
	expect(ll_uint64_eq(lzma_index_total_size(i), total_size));
	expect(ll_uint64_eq(lzma_index_uncompressed_size(i), uncompressed_size));
	expect(ll_uint64_eq(ll_uint64_add(ll_uint64_add(lzma_index_total_size(i), lzma_index_size(i)), ll_uint64_make(0, 2 * LZMA_STREAM_HEADER_SIZE)),
		lzma_index_stream_size(i)));

	return i;
}


static bool is_equal(const lzma_index *a, const lzma_index *b)
{
	/* Compare only the Stream and Block sizes and offsets. */
	lzma_index_iter ra, rb;

	lzma_index_iter_init(&ra, a);
	lzma_index_iter_init(&rb, b);

	for (;;)
	{
		bool reta = lzma_index_iter_next(&ra, LZMA_INDEX_ITER_ANY);
		bool retb = lzma_index_iter_next(&rb, LZMA_INDEX_ITER_ANY);

		if (reta)
			return !(reta ^ retb);

		if (!ll_uint64_eq(ra.stream.number, rb.stream.number) ||
			!ll_uint64_eq(ra.stream.block_count, rb.stream.block_count) ||
			!ll_uint64_eq(ra.stream.compressed_offset, rb.stream.compressed_offset) ||
			!ll_uint64_eq(ra.stream.uncompressed_offset, rb.stream.uncompressed_offset) ||
			!ll_uint64_eq(ra.stream.compressed_size, rb.stream.compressed_size) ||
			!ll_uint64_eq(ra.stream.uncompressed_size, rb.stream.uncompressed_size) ||
			!ll_uint64_eq(ra.stream.padding, rb.stream.padding))
			return false;

		if (ll_uint64_iszero(ra.stream.block_count))
			continue;

		if (!ll_uint64_eq(ra.block.number_in_file, rb.block.number_in_file) ||
			!ll_uint64_eq(ra.block.compressed_file_offset, rb.block.compressed_file_offset) ||
			!ll_uint64_eq(ra.block.uncompressed_file_offset, rb.block.uncompressed_file_offset) ||
			!ll_uint64_eq(ra.block.number_in_stream, rb.block.number_in_stream) ||
			!ll_uint64_eq(ra.block.compressed_stream_offset, rb.block.compressed_stream_offset) ||
			!ll_uint64_eq(ra.block.uncompressed_stream_offset, rb.block.uncompressed_stream_offset) ||
			!ll_uint64_eq(ra.block.uncompressed_size, rb.block.uncompressed_size) ||
			!ll_uint64_eq(ra.block.unpadded_size, rb.block.unpadded_size) ||
			!ll_uint64_eq(ra.block.total_size, rb.block.total_size))
			return false;
	}
}


static void test_equal(void)
{
	lzma_index *a = create_empty();
	lzma_index *b = create_small();
	lzma_index *c = create_big();

	expect(a && b && c);

	expect(is_equal(a, a));
	expect(is_equal(b, b));
	expect(is_equal(c, c));

	expect(!is_equal(a, b));
	expect(!is_equal(a, c));
	expect(!is_equal(b, c));

	lzma_index_end(a, NULL);
	lzma_index_end(b, NULL);
	lzma_index_end(c, NULL);
}


static void test_overflow(void)
{
	/* Integer overflow tests */
	lzma_index *i = create_empty();

	expect(lzma_index_append(i, NULL, ll_uint64_sub(LZMA_VLI_MAX, ll_uint64_make(0, 5)), ll_uint64_make(0, 1234)) == LZMA_DATA_ERROR);

	/* TODO */

	lzma_index_end(i, NULL);
}


static void test_copy(const lzma_index *i)
{
	lzma_index *d = lzma_index_dup(i, NULL);

	expect(d != NULL);
	expect(is_equal(i, d));
	lzma_index_end(d, NULL);
}


static void test_read(lzma_index *i)
{
	lzma_index_iter r;
	size_t j;

	lzma_index_iter_init(&r, i);

	/* Try twice so we see that rewinding works. */
	for (j = 0; j < 2; ++j)
	{
		lzma_vli total_size = ll_uint64_zero;
		lzma_vli uncompressed_size = ll_uint64_zero;
		lzma_vli stream_offset = ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE);
		lzma_vli uncompressed_offset = ll_uint64_zero;
		uint32_t count = 0;

		while (!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK))
		{
			++count;

			total_size = ll_uint64_add(total_size, r.block.total_size);
			uncompressed_size = ll_uint64_add(uncompressed_size, r.block.uncompressed_size);

			expect(ll_uint64_eq(r.block.compressed_file_offset, stream_offset));
			expect(ll_uint64_eq(r.block.uncompressed_file_offset, uncompressed_offset));

			stream_offset = ll_uint64_add(stream_offset, r.block.total_size);
			uncompressed_offset = ll_uint64_add(uncompressed_offset, r.block.uncompressed_size);
		}

		expect(ll_uint64_eq(lzma_index_total_size(i), total_size));
		expect(ll_uint64_eq(lzma_index_uncompressed_size(i), uncompressed_size));
		expect(ll_uint64_eq(lzma_index_block_count(i), ll_uint64_make(0, count)));

		lzma_index_iter_rewind(&r);
	}
}


static void test_code(lzma_index *i)
{
	const size_t alloc_size = UINT32_C(0x20000);
	uint8_t *buf = malloc(alloc_size);
	size_t buf_pos;
	size_t pos;
	lzma_mem_t memlimit;
	lzma_index_iter r;
	lzma_index_hash *h;
	lzma_index *d;
	lzma_vli index_size;
	lzma_stream strm = LZMA_STREAM_INIT;

	expect(buf != NULL);

	/* Encode */
	expect(lzma_index_encoder(&strm, i) == LZMA_OK);
	index_size = lzma_index_size(i);
	succeed(coder_loop(&strm, NULL, 0, buf, ll_uint64_to_size_t(index_size), LZMA_STREAM_END, LZMA_RUN));

	/* Decode */
	expect(lzma_index_decoder(&strm, &d, ll_uint64_to_size_t(MEMLIMIT)) == LZMA_OK);
	expect(d == NULL);
	succeed(decoder_loop(&strm, buf, ll_uint64_to_size_t(index_size)));

	expect(is_equal(i, d));

	lzma_index_end(d, NULL);
	lzma_end(&strm);

	/* Decode with hashing */
	h = lzma_index_hash_init(NULL, NULL);
	expect(h != NULL);
	lzma_index_iter_init(&r, i);
	while (!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK))
		expect(lzma_index_hash_append(h, r.block.unpadded_size, r.block.uncompressed_size) == LZMA_OK);
	pos = 0;
	while (pos < ll_uint64_to_size_t(index_size) - 1)
		expect(lzma_index_hash_decode(h, buf, &pos, pos + 1) == LZMA_OK);
	expect(lzma_index_hash_decode(h, buf, &pos, pos + 1) == LZMA_STREAM_END);

	lzma_index_hash_end(h, NULL);

	/* Encode buffer */
	buf_pos = 1;
	expect(lzma_index_buffer_encode(i, buf, &buf_pos, ll_uint64_to_size_t(index_size)) == LZMA_BUF_ERROR);
	expect(buf_pos == 1);

	succeed(lzma_index_buffer_encode(i, buf, &buf_pos, ll_uint64_to_size_t(index_size) + 1));
	expect(buf_pos == ll_uint64_to_size_t(index_size) + 1);

	/* Decode buffer */
	buf_pos = 1;
	memlimit = ll_uint64_to_size_t(MEMLIMIT);
	expect(lzma_index_buffer_decode(&d, &memlimit, NULL, buf, &buf_pos, ll_uint64_to_size_t(index_size)) == LZMA_DATA_ERROR);
	expect(buf_pos == 1);
	expect(d == NULL);

	succeed(lzma_index_buffer_decode(&d, &memlimit, NULL, buf, &buf_pos, ll_uint64_to_size_t(index_size) + 1));
	expect(buf_pos == ll_uint64_to_size_t(index_size) + 1);
	expect(is_equal(i, d));

	lzma_index_end(d, NULL);

	free(buf);
}


static void test_many(lzma_index *i)
{
	test_copy(i);
	test_read(i);
	test_code(i);
}


static void test_cat(void)
{
	lzma_index *a, *b, *c;
	lzma_index_iter r;
	lzma_vli stream_size;
	int i;

	/* Empty Indexes */
	a = create_empty();
	b = create_empty();
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_block_count(a), ll_uint64_zero));
	expect(ll_uint64_eq(lzma_index_stream_size(a), ll_uint64_make(0, 2 * LZMA_STREAM_HEADER_SIZE + 8)));
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_make(0, 2 * (2 * LZMA_STREAM_HEADER_SIZE + 8))));
	lzma_index_iter_init(&r, a);
	expect(lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK));

	b = create_empty();
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_block_count(a), ll_uint64_zero));
	expect(ll_uint64_eq(lzma_index_stream_size(a), ll_uint64_make(0, 2 * LZMA_STREAM_HEADER_SIZE + 8)));
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_make(0, 3 * (2 * LZMA_STREAM_HEADER_SIZE + 8))));

	b = create_empty();
	c = create_empty();
	expect(lzma_index_stream_padding(b, ll_uint64_make(0, 4)) == LZMA_OK);
	expect(lzma_index_cat(b, c, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_block_count(b), ll_uint64_zero));
	expect(ll_uint64_eq(lzma_index_stream_size(b), ll_uint64_make(0, 2 * LZMA_STREAM_HEADER_SIZE + 8)));
	expect(ll_uint64_eq(lzma_index_file_size(b), ll_uint64_make(0, 2 * (2 * LZMA_STREAM_HEADER_SIZE + 8) + 4)));

	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 8)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_block_count(a), ll_uint64_zero));
	expect(ll_uint64_eq(lzma_index_stream_size(a), ll_uint64_make(0, 2 * LZMA_STREAM_HEADER_SIZE + 8)));
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_make(0, 5 * (2 * LZMA_STREAM_HEADER_SIZE + 8) + 4 + 8)));

	expect(lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK));
	lzma_index_iter_rewind(&r);
	expect(lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK));
	lzma_index_end(a, NULL);

	/* Small Indexes */
	a = create_small();
	stream_size = lzma_index_stream_size(a);
	lzma_index_iter_init(&r, a);
	for (i = SMALL_COUNT; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	b = create_small();
	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 4)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_add(ll_uint64_lshift(stream_size, 1), ll_uint64_make(0, 4))));
	expect(ll_uint64_gt(lzma_index_stream_size(a), stream_size));
	expect(ll_uint64_lt(lzma_index_stream_size(a), ll_uint64_lshift(stream_size, 1)));
	for (i = SMALL_COUNT; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	lzma_index_iter_rewind(&r);
	for (i = SMALL_COUNT * 2; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	b = create_small();
	c = create_small();
	expect(lzma_index_stream_padding(b, ll_uint64_make(0, 8)) == LZMA_OK);
	expect(lzma_index_cat(b, c, NULL) == LZMA_OK);
	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 12)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_add(ll_uint64_lshift(stream_size, 2), ll_uint64_make(0, 4 + 8 + 12))));

	expect(ll_uint64_eq(lzma_index_block_count(a), ll_uint64_make(0, SMALL_COUNT * 4)));
	for (i = SMALL_COUNT * 2; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	lzma_index_iter_rewind(&r);
	for (i = SMALL_COUNT * 4; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	lzma_index_end(a, NULL);

	/* Mix of empty and small */
	a = create_empty();
	b = create_small();
	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 4)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	lzma_index_iter_init(&r, a);
	for (i = SMALL_COUNT; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	lzma_index_end(a, NULL);

	/* Big Indexes */
	a = create_big();
	stream_size = lzma_index_stream_size(a);
	b = create_big();
	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 4)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_add(ll_uint64_lshift(stream_size, 1), ll_uint64_make(0, 4))));
	expect(ll_uint64_gt(lzma_index_stream_size(a), stream_size));
	expect(ll_uint64_lt(lzma_index_stream_size(a), ll_uint64_lshift(stream_size, 1)));

	b = create_big();
	c = create_big();
	expect(lzma_index_stream_padding(b, ll_uint64_make(0, 8)) == LZMA_OK);
	expect(lzma_index_cat(b, c, NULL) == LZMA_OK);
	expect(lzma_index_stream_padding(a, ll_uint64_make(0, 12)) == LZMA_OK);
	expect(lzma_index_cat(a, b, NULL) == LZMA_OK);
	expect(ll_uint64_eq(lzma_index_file_size(a), ll_uint64_add(ll_uint64_lshift(stream_size, 2), ll_uint64_make(0, 4 + 8 + 12))));

	lzma_index_iter_init(&r, a);
	for (i = BIG_COUNT * 4; i >= 0; --i)
		expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK) ^ (i == 0));

	lzma_index_end(a, NULL);
}


static void test_locate(void)
{
	lzma_index *i = lzma_index_init(NULL);
	lzma_index_iter r;
	size_t group_multiple;
	size_t radius;
	size_t start;
	lzma_vli ubase;
	lzma_vli tbase;
	size_t n;

	expect(i != NULL);
	lzma_index_iter_init(&r, i);

	/* Cannot locate anything from an empty Index. */
	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));
	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 555)));

	/* One empty Record: nothing is found since there's no uncompressed */
	/* data. */
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 16), ll_uint64_make(0, 0)) == LZMA_OK);
	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));

	/* Non-empty Record and we can find something. */
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 32), ll_uint64_make(0, 5)) == LZMA_OK);
	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 32)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 5)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	/* Still cannot find anything past the end. */
	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 5)));

	/* Add the third Record. */
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 40), ll_uint64_make(0, 11)) == LZMA_OK);

	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 32)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 5)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	expect(!lzma_index_iter_next(&r, LZMA_INDEX_ITER_BLOCK));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 40)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 11)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16 + 32)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 5)));

	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 2)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 32)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 5)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 5)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 40)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 11)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16 + 32)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 5)));

	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 5 + 11 - 1)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 40)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 11)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 16 + 32)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 5)));

	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 5 + 11)));
	expect(lzma_index_iter_locate(&r, ll_uint64_make(0, 5 + 15)));

	/* Large Index */
	lzma_index_end(i, NULL);
	i = lzma_index_init(NULL);
	expect(i != NULL);
	lzma_index_iter_init(&r, i);

	for (n = 4; n <= 4 * 5555; n += 4)
		expect(lzma_index_append(i, NULL, ll_uint64_make(0, n + 8), ll_uint64_make(0, n)) == LZMA_OK);

	expect(ll_uint64_eq(lzma_index_block_count(i), ll_uint64_make(0, 5555)));

	/* First Record */
	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 4 + 8)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 4)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 3)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 4 + 8)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 4)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	/* Second Record */
	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 4)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 2 * 4 + 8)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 2 * 4)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + 4 + 8)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 4)));

	/* Last Record */
	expect(!lzma_index_iter_locate(&r, ll_uint64_sub(lzma_index_uncompressed_size(i), ll_uint64_make(0, 1))));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 4 * 5555 + 8)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 4 * 5555)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_sub(ll_uint64_add(lzma_index_total_size(i), ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE)), ll_uint64_make(0, 4 * 5555 + 8))));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_sub(lzma_index_uncompressed_size(i), ll_uint64_make(0, 4 * 5555))));

	/* Allocation chunk boundaries. See INDEX_GROUP_SIZE in */
	/* liblzma/common/index.c. */
	group_multiple = 256 * 4;
	radius = 8;
	start = group_multiple - radius;
	ubase = ll_uint64_zero;
	tbase = ll_uint64_zero;
	for (n = 1; n < start; ++n)
	{
		ubase = ll_uint64_add(ubase, ll_uint64_make(0, n * 4));
		tbase = ll_uint64_add(tbase, ll_uint64_make(0, n * 4 + 8));
	}

	while (n < start + 2 * radius)
	{
		expect(!lzma_index_iter_locate(&r, ll_uint64_add(ubase, ll_uint64_make(0, n * 4))));

		expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_add(tbase, ll_uint64_make(0, n * 4 + 8 + LZMA_STREAM_HEADER_SIZE))));
		expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_add(ubase, ll_uint64_make(0, n * 4))));

		tbase = ll_uint64_add(tbase, ll_uint64_make(0, n * 4 + 8));
		ubase = ll_uint64_add(ubase, ll_uint64_make(0, n * 4));
		++n;

		expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, n * 4 + 8)));
		expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, n * 4)));
	}

	/* Do it also backwards. */
	while (n > start)
	{
		expect(!lzma_index_iter_locate(&r, ll_uint64_add(ubase, ll_uint64_make(0, (n - 1) * 4))));

		expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, n * 4 + 8)));
		expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, n * 4)));

		--n;
		tbase = ll_uint64_sub(tbase, ll_uint64_make(0, n * 4 + 8));
		ubase = ll_uint64_sub(ubase, ll_uint64_make(0, n * 4));

		expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_add(tbase, ll_uint64_make(0, n * 4 + 8 + LZMA_STREAM_HEADER_SIZE))));
		expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_add(ubase, ll_uint64_make(0, n * 4))));
	}

	/* Test locating in concatenated Index. */
	lzma_index_end(i, NULL);
	i = lzma_index_init(NULL);
	expect(i != NULL);
	lzma_index_iter_init(&r, i);
	for (n = 0; n < group_multiple; ++n)
		expect(lzma_index_append(i, NULL, ll_uint64_make(0, 8), ll_uint64_make(0, 0)) == LZMA_OK);
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 16), ll_uint64_make(0, 1)) == LZMA_OK);
	expect(!lzma_index_iter_locate(&r, ll_uint64_make(0, 0)));
	expect(ll_uint64_eq(r.block.total_size, ll_uint64_make(0, 16)));
	expect(ll_uint64_eq(r.block.uncompressed_size, ll_uint64_make(0, 1)));
	expect(ll_uint64_eq(r.block.compressed_file_offset, ll_uint64_make(0, LZMA_STREAM_HEADER_SIZE + group_multiple * 8)));
	expect(ll_uint64_eq(r.block.uncompressed_file_offset, ll_uint64_make(0, 0)));

	lzma_index_end(i, NULL);
}


static void test_corrupt(void)
{
	const size_t alloc_size = UINT32_C(0x20000);
	uint8_t *buf = malloc(alloc_size);
	lzma_stream strm = LZMA_STREAM_INIT;
	lzma_index *i;

	expect(buf != NULL);

	i = create_empty();
	expect(lzma_index_append(i, NULL, ll_uint64_make(0, 0), ll_uint64_make(0, 1)) == LZMA_PROG_ERROR);
	lzma_index_end(i, NULL);

	/* Create a valid Index and corrupt it in different ways. */
	i = create_small();
	expect(lzma_index_encoder(&strm, i) == LZMA_OK);
	succeed(coder_loop(&strm, NULL, 0, buf, 20, LZMA_STREAM_END, LZMA_RUN));
	lzma_index_end(i, NULL);

	/* Wrong Index Indicator */
	buf[0] ^= 1;
	expect(lzma_index_decoder(&strm, &i, ll_uint64_to_size_t(MEMLIMIT)) == LZMA_OK);
	succeed(decoder_loop_ret(&strm, buf, 1, LZMA_DATA_ERROR));
	buf[0] ^= 1;

	/* Wrong Number of Records and thus CRC32 fails. */
	--buf[1];
	expect(lzma_index_decoder(&strm, &i, ll_uint64_to_size_t(MEMLIMIT)) == LZMA_OK);
	succeed(decoder_loop_ret(&strm, buf, 10, LZMA_DATA_ERROR));
	++buf[1];

	/* Padding not NULs */
	buf[15] ^= 1;
	expect(lzma_index_decoder(&strm, &i, ll_uint64_to_size_t(MEMLIMIT)) == LZMA_OK);
	succeed(decoder_loop_ret(&strm, buf, 16, LZMA_DATA_ERROR));

	lzma_end(&strm);
	free(buf);
}


/* Allocator that succeeds for the first two allocation but fails the rest. */
static void *LZMA_API_CALL my_alloc(void *opaque, size_t a, size_t b)
{
	static unsigned count = 0;

	(void) opaque;
	if (++count > 2)
		return NULL;

	return malloc(a * b);
}

static const lzma_allocator my_allocator = { my_alloc, NULL, NULL };


int main(void)
{
	lzma_index *i;

#if defined(LZMA_SLB) && (defined(__atarist__) || defined(__TOS__))
	{
		long ret;

		if ((ret = slb_lzma_open("../atari")) < 0)
		{
			fprintf(stderr, "cannot load " LZMA_SHAREDLIB_NAME ": %ld\n", ret);
			return 1;
		}
	}
#endif
	test_equal();

	test_overflow();

	i = create_empty();
	test_many(i);
	lzma_index_end(i, NULL);

	i = create_small();
	test_many(i);
	lzma_index_end(i, NULL);

	i = create_big();
	test_many(i);
	lzma_index_end(i, NULL);

	test_cat();

	test_locate();

	test_corrupt();

	/* Test for the bug fix 21515d79d778b8730a434f151b07202d52a04611: */
	/* liblzma: Fix lzma_index_dup() for empty Streams. */
	i = create_empty();
	expect(lzma_index_stream_padding(i, ll_uint64_make(0, 4)) == LZMA_OK);
	test_copy(i);
	lzma_index_end(i, NULL);

	/*
	 * Test for the bug fix 3bf857edfef51374f6f3fffae3d817f57d3264a0:
	 * liblzma: Fix a memory leak in error path of lzma_index_dup().
	 * Use Valgrind to see that there are no leaks.
	 */
	i = create_small();
	expect(lzma_index_dup(i, &my_allocator) == NULL);
	lzma_index_end(i, NULL);

	return 0;
}
